from ..common import messages as m
import asyncio
import logging
import contextlib
from .banner import banner

class ClientInfo():
    def __init__(self,ttyId):
        self.ttyId = ttyId
class FiniteQueue(asyncio.queues.Queue):
    End = object()
    def __aiter__(self):
        return self
    async def close(self):
        await self.put(FiniteQueue.End) # how to dispatch this to all getters?
    async def __anext__(self):
        elt = await self.get()
        if elt is FiniteQueue.End:
            # maybe here? self.put_nowait(FiniteQueue.End)
            raise StopAsyncIteration
        else:
            return elt


class BaseClient:
    def __init__(self, path):
        self.sendq = FiniteQueue()
        self.path = path

    async def run(self,socket):
        self.socket = socket
        log.info('{} connected'.format(self.path))
        self.psq = asyncio.ensure_future(self.process_sendq())

        await self.first_messages()

        # wait until the other side closes the socket properly
        with contextlib.suppress(asyncio.CancelledError):  # or the connection is dropped
            async for msg in self.socket:
                handled = await self.handle_message(msg)
                if not handled:
                    log.debug('{} > [IGNORED] {}'.format(self.path, msg.data))

        # put this in a with:__exit__ instead of suppressing CancelledError?
        self.psq.cancel()

    async def send_text(self,text):
        await self.sendq.put(text)

    async def first_messages(self):
        pass

    async def handle_message(self,msg):
        pass

    async def close(self):
        log.info('{} < [CLOSE]'.format(self.path))
        await self.socket.close()

    async def process_sendq(self):
        async for msg in self.sendq:
            log.debug('{} < {}'.format(self.path, msg))
            await self.socket.send_str(msg)
        await self.socket.close()


class FeedClient(BaseClient):
    def __init__(self,path,ttyId):
        super().__init__(path)
        self.info = ClientInfo(ttyId)
        self.rlc = None
        self.devc = None

    async def first_messages(self):
        await self.send_banner()

    async def handle_message(self, raw):
        try:
            msg = m.Message.parse(raw.data)
        except:
            msg = None
        if isinstance(msg, m.Info):
            for k, v in msg.__dict__.items():
                setattr(self.info, k, v)
            log.debug('{} > INFO({})'.format(self.path, self.info.__dict__))
            await self.send_passphrase()
        elif isinstance(msg,m.LineRead) and self.rlc:
            await self.rlc.send(msg.text)
        elif isinstance(msg, m.KeyRead) and self.rlc:
            await self.rlc.send(msg.key)
        else:
            return False # message ignored
        return True #message handled

    async def send(self,cmd):
        assert isinstance(cmd, m.Message)
        await self.send_text(str(cmd))

    async def send_banner(self):
        for line in banner.splitlines():
            await self.send(m.Line(line))

    async def send_passphrase(self):
        text = 'Access this terminal with the following passphrase:'
        await self.send(m.Line(text))
        await self.send(m.Line(self.info.passphrase))

    async def read_line(self,path,socket):
        self.rlc = ReadLineClient(path,self)
        await self.send(m.ReadLine())
        await self.rlc.run(socket)
        await self.rlc.close()
        self.rlc = None

    async def read_key(self,path,socket,echo):
        self.rlc = ReadLineClient(path,self)
        await self.send(m.ReadKey(echo=echo))
        await self.rlc.run(socket)
        await self.rlc.close()
        self.rlc = None

    async def run_dev_client(self,path,socket):
        self.devc = DevClient(path,self)
        await self.devc.run(socket)
        await self.devc.close()
        self.devc = None

    async def close(self):
        await super().close()
        if self.rlc:
            await self.rlc.close()
        if self.devc:
            await self.devc.close()

class DevClient(BaseClient):
    def __init__(self, path, yinClient):
        super().__init__(path)
        self.yinClient = yinClient
        self.info = ClientInfo(yinClient.info.ttyId)

    async def handle_message(self,msg):
        log.debug('{} > {}'.format(self.info.path, msg.data))
        await self.yinClient.send_text(msg.data)
        return True

class ReadLineClient(BaseClient):
    def __init__(self, path, yinClient):
        super().__init__(path)
        self.info = ClientInfo(yinClient.info.ttyId)

    async def send(self, text):
        await self.send_text(text)
        await self.send_text(FiniteQueue.End)


log = logging.getLogger(__name__)



import asyncio.queues
import logging
from aiohttp import web

from .routes import routes
from .state import State
from .exceptions import AbortRequestException

def create():
    app = web.Application(middlewares=[cors_header, exception_mapper])
    State.init(app)
    app.router.add_routes(routes)
    app.on_shutdown.append(close_websockets)
    return app


# async def shutdown_close(ws):
#    await ws.close(code=http.WSCloseCode.GOING_AWAY, message='Server shutdown')
#    log.debug('socket status: {}:{}'.format(ws.closed,ws.close_code))

async def close_websockets(app):
    client_closers = [c.close() for c in State.of(app).clients.values()]
    if len(client_closers)>0:
        await asyncio.wait(client_closers)
    log.debug('closed {} clients'.format(len(client_closers)))


@web.middleware
async def cors_header(request,handler):
    resp = await handler(request)
    resp.headers['Access-Control-Allow-Origin'] = '*'
    return resp

@web.middleware
async def exception_mapper(request,handler):
    try:
        resp = await handler(request)
        return resp
    except AbortRequestException as e:
        return web.Response(status=e.status, text=e.text)


log = logging.getLogger(__name__)




from .conf import ServerConfig
import asyncio
import logging
import warnings
import aiohttp.web
from . import app
import sys

def main():
    if len(sys.argv)>1:
        ServerConfig.read(sys.argv[1])

    print(type(ServerConfig.debug))
    print(type(ServerConfig.logfile))
    print(type(ServerConfig.port))

    if ServerConfig.debug:
        asyncio.get_event_loop().set_debug(True)
        warnings.simplefilter("always", ResourceWarning)
    logging.basicConfig(level=logging.DEBUG if ServerConfig.debug else logging.INFO, filename=ServerConfig.logfile)
    aiohttp.web.run_app(app.create(),port=ServerConfig.port)

if __name__=='__main__':
    main()


import yaml
import logging

class ServerConfig:
    port = 8080
    logfile = None  # 'server.log'
    debug = True

    @staticmethod
    def read(filename):
        with open(filename) as f:
            yml = yaml.safe_load(f)
        for key,val in yml.items():
            if hasattr(ServerConfig,key):
                setattr(ServerConfig,key,val)
            else:
                logging.warning('Unknown configuration key "{}"'.format(key))



from .exceptions import AbortRequestException
from .state import State
from ..common import messages as m
from ..common import id_generator
from .clients import FeedClient
from aiohttp import web

routes = web.RouteTableDef()

def augment(handler):
    def augmented(req):
        info = get_req_info(req)
        return handler(req,**info)
    return augmented

@routes.get('/tty/id')
@augment
async def find_id(req, path):
    try:
        words = req.query.getall('phrase')
    except KeyError:
        raise AbortRequestException(status=400, text='Usage: GET {}?phrase=example%20pass%20phrase\n'.format(path))
    ttyId = id_generator.id_hash(words)
    return web.Response(text=ttyId)

@routes.get('/tty/{ttyId}/dev')
@augment
async def dev_endpoint(req, path, ttyId):
    client = find_client(req.app, ttyId)
    if client.devc:
        return web.Response(status=409, text='This endpoint is already in use.')
    socket = web.WebSocketResponse(heartbeat=15)
    await socket.prepare(req)
    await client.run_dev_client(path,socket)
    return socket

@routes.get('/tty/{ttyId}/readline')
@augment
async def read_line(req, path, ttyId):
    client = find_client(req.app, ttyId)
    if client.rlc:
        return web.Response(status=409, text='This endpoint is already in use.')
    socket = web.WebSocketResponse(heartbeat=15)
    await socket.prepare(req)
    await client.read_line(path,socket)
    return socket

@routes.get('/tty/{ttyId}/readkey')
@augment
async def read_line(req, path, ttyId):
    client = find_client(req.app, ttyId)
    if client.rlc:
        return web.Response(status=409, text='This endpoint is already in use.')
    echo = parse_echo(req.query.getall('echo',[]))
    socket = web.WebSocketResponse(heartbeat=15)
    await socket.prepare(req)
    await client.read_key(path,socket,echo)
    return socket

def parse_echo(queryvals):
    return 'on' in queryvals

@routes.get('/feed/{ttyId}')
@augment
async def feed_endpoint(req, path, ttyId):
    clients = State.of(req.app).clients
    if ttyId in clients:
        return web.Response(status=409, text='This endpoint is already in use.')
    client = FeedClient(path, ttyId)
    clients[ttyId] = client

    socket = web.WebSocketResponse(heartbeat=15)
    await socket.prepare(req)
    await client.run(socket)
    await client.close()
    del clients[ttyId]
    return socket

@routes.post('/tty/{ttyId}/line')
@augment
async def post_line(req, path, ttyId):
    client = find_client(req.app, ttyId)
    text = await req.text()
    opts = parse_align_opts(req.query.getall('align',[]),'h')

    await client.send(m.Line(text,**opts))
    return web.Response(status=200)

@routes.post('/tty/{ttyId}/page')
@augment
async def post_page(req, path, ttyId):
    client = find_client(req.app, ttyId)
    text = await req.text()
    opts = parse_align_opts(req.query.getall('align',[]),'hv')

    await client.send(m.Page(text,**opts))
    return web.Response(status=200)

def parse_align_opts(queryvals,hv):
    vals = [v for qv in queryvals for v in qv.split()]
    halign = [val for val in vals if val in ('left','center','right')]
    valign = [val for val in vals if val in ('top','middle','bottom')]
    opts = {}
    if halign and 'h' in hv:
        opts['halign'] = halign[0]
    if valign and 'v' in hv:
        opts['valign'] = valign[0]
    return opts

@routes.post('/tty/{ttyId}/cls')
@augment
async def post_cls(req, path, ttyId):
    client = find_client(req.app, ttyId)
    await client.send(m.Cls())
    return web.Response(status=200)

def get_req_info(req):
    info = dict(req.match_info)
    info['path'] = req.match_info.route.url_for(**req.match_info)
    return info

def find_client(app, ttyId):
    try:
        return State.of(app).clients[ttyId]
    except KeyError:
        raise AbortRequestException(status=403, text='magnie')




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='run80by24-server',
    version='0.1.0',
    description='Server for publishing to text terminals over HTTP.',
    long_description=long_description,
    url='https://80by24.net',
    author='Sander Evers',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='terminal console',
    packages=['run80by24.'+pkg for pkg in find_packages('run80by24')],
    install_requires=['aiohttp>=2.3.2','pyyaml','run80by24-common'],  # Optional
    # package_data={  # Optional
    #     'sample': ['package_data.dat'],
    # },

    entry_points={
        'console_scripts': [
            'serve_80by24=run80by24.server.__main__:main',
        ],
    },
)